@@ -1,10 +1,11 @@
 import struct
-from mmap import mmap
 
 from typing import Union, List, Dict, Tuple
 
 from DyldExtractor.file_context import FileContext
 
+from DyldExtractor.dyld.dyld_structs import dyld_cache_mapping_info
+
 from DyldExtractor.macho.segment_context import SegmentContext
 from DyldExtractor.macho.macho_structs import (
 	LoadCommandMap,
@@ -16,7 +17,7 @@ from DyldExtractor.macho.macho_structs import (
 )
 
 
-class MachOContext(FileContext):
+class MachOContext(object):
 
 	loadCommands: List[load_command]
 
@@ -25,7 +26,7 @@ class MachOContext(FileContext):
 
 	def __init__(
 		self,
-		file: mmap,
+		fileCtx: FileContext,
 		offset: int
 	) -> None:
 		"""A wrapper around a MachO file.
@@ -37,9 +38,13 @@ class MachOContext(FileContext):
 			offset: The offset to the header in the file.
 		"""
 
-		super().__init__(file, offset=offset)
+		super().__init__()
 
-		self.header = mach_header_64(file, offset)
+		self.fileCtx = fileCtx
+		self.fileOffset = offset
+
+		self.header = mach_header_64(self.fileCtx.file, self.fileOffset)
+		self._mappings: List[Tuple[dyld_cache_mapping_info, FileContext]] = []
 
 		# check to make sure the MachO file is 64 bit
 		magic = self.header.magic
@@ -105,30 +110,83 @@ class MachOContext(FileContext):
 
 		Parse the load commands and set the loadCommands attribute.
 		"""
+		self.header = mach_header_64(self.fileCtx.file, self.fileOffset)
 
 		self.loadCommands = []
 
 		self.segments = {}
 		self.segmentsI = []
 
+		file = self.fileCtx.file
 		cmdOff = len(self.header) + self.fileOffset
 		for _ in range(self.header.ncmds):
-			self.file.seek(cmdOff)
-			cmd = struct.unpack("<I", self.file.read(4))[0]
+			file.seek(cmdOff)
+			cmd = struct.unpack("<I", file.read(4))[0]
 
 			command = LoadCommandMap.get(cmd, UnknownLoadCommand)
 			if command == UnknownLoadCommand:
 				raise Exception(f"Unknown LoadCommand: {cmd}")
 
-			command = command(self.file, cmdOff)
+			command = command(file, cmdOff)
 
 			cmdOff += command.cmdsize
 			self.loadCommands.append(command)
 
 			# populate the segments at this point too
 			if isinstance(command, segment_command_64):
-				segCtx = SegmentContext(self.file, command)
+				segCtx = SegmentContext(file, command)
 
 				self.segments[command.segname] = segCtx
 				self.segmentsI.append(segCtx)
+				pass
+			pass
 		pass
+
+	def reloadLoadCommands(self) -> None:
+		"""Reload the load commands.
+
+		Read the header and load commands with the
+		same file and offset.
+		"""
+
+		self._parseLoadCommands()
+		pass
+
+	def addSubfiles(
+		self,
+		mainFileMap: dyld_cache_mapping_info,
+		subFilesAndMaps: List[Tuple[dyld_cache_mapping_info, FileContext]]
+	) -> None:
+		"""Add sub files.
+
+		Used when the data to describe a MachO file is split into multiple
+		files, each file needs to have mapping info.
+
+		Args:
+			mainFileMap: Mapping info for the file that contains the header.
+			subFilesAndMaps: A list of tuples that contain the sub file context
+				and the mapping info.
+		"""
+		self._mappings.append((mainFileMap, self.fileCtx))
+		self._mappings.extend(subFilesAndMaps)
+		pass
+
+	def fileForAddr(self, vmaddr: int) -> FileContext:
+		"""Get the file context that contains the address.
+
+		If there are no sub files added, this just returns the main
+		file context. If the file cannot be found, this returns None.
+		"""
+
+		if not self._mappings:
+			return self.fileCtx
+
+		for mapping, ctx in self._mappings:
+			lowBound = mapping.address
+			highBound = mapping.address + mapping.size
+
+			if vmaddr >= lowBound and vmaddr < highBound:
+				return ctx
+
+		# didn't find the address in any mappings...
+		return None
